/*
    Ishtar - A network transparent read/write interface with
    probing and enumeration support.
    Copyright (C) 2003-2005 Stephane Magnenat <nct@ysagoon.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <ishtar/services.h>

using namespace Ishtar;

struct Request
{
	unsigned counter;
	ValuesAnswer answer;
	Socket *client;

	Request(UInt32 count, Socket *c) { counter = count; client=c; }
};

class Proxyd : public NetworkServer
{
	//! Type that identify a service on this host. Pointer is server, int is service
	typedef std::pair<Socket *, UInt32> InternalServiceID;

	//! Type that map external service to internal one
	typedef std::map<UInt32, InternalServiceID> ExternalServiceMap;

	//! Type that map internal service to external one
	typedef std::map<InternalServiceID, UInt32> InternalServiceMap;

	//! Type that map server ID to service description
	typedef std::map<Socket *, ServiceAnswer *> ServerDescriptionMap;

	//! Type that map request ID to request
	typedef std::map<UInt32, Request *> RequestMap;

	//! Map of external services vs internal. Used for routing
	ExternalServiceMap externalServiceMap;

	//! Map of internal services vs external. Used for routing
	InternalServiceMap internalServiceMap;

	//! Map of server's ID to service description
	ServerDescriptionMap serverDescriptionMap;

	//! List of all pending requests
	RequestMap requests;

	//! The service id for external access is dynamically allocated, we are sure each value is only used once
	unsigned freeServiceId;
	//! The request id is dynamically allocated, we are sure each value is only used once
	unsigned freeRequestId;

public:
	Proxyd(UInt16 port)
	:NetworkServer(port)
	{
		freeServiceId = 0;
		freeRequestId = 0;
	}

	UInt32 getGlobalID(Socket *s, UInt32 id)
	{
		InternalServiceID isid(s, id);
		return internalServiceMap[isid];
	}

	void incomingData(Socket *socket)
	{
		std::cout << "Incoming data" << std::endl;
		ServerDescriptionMap::const_iterator serverIt = serverDescriptionMap.find(socket);
		if (serverIt != serverDescriptionMap.end())
		{
			// This is data from the servers, so it's an answer
			Answer *answer = Answer::deserialize(socket);
			if (socket->isConnected() == false)
				return;

			ServiceAnswer *serviceAnswer = dynamic_cast<ServiceAnswer *>(answer);
			if (serviceAnswer)
			{
				// We got a ServicesAnswer from one of our server
				// If we have already one, delete it
				if (serverDescriptionMap[socket])
				{
					serverDescriptionMap[socket]->deleteDescriptions();
					delete serverDescriptionMap[socket];
				}
				serverDescriptionMap[socket] = serviceAnswer;

				// Create association of services description
				for (ServiceAnswer::DescriptionMap::const_iterator i=serviceAnswer->descriptions.begin(); i!=serviceAnswer->descriptions.end(); ++i)
				{
					std::string name = i->second->name;
					std::cout << "Adding service " << name << std::endl;
					InternalServiceID isid(socket, i->first);
					externalServiceMap[freeServiceId] = isid;
					internalServiceMap[isid] = freeServiceId;
					freeServiceId++;
				}

				return;
			}

			ValuesAnswer *valuesAnswer = dynamic_cast<ValuesAnswer *>(answer);
			if (valuesAnswer)
			{
				// We read request ID, then lookup in request pending table
				UInt32 requestID = valuesAnswer->requestID;
				Request *r = requests[requestID];

				for (ValuesAnswer::ValueVectorMap::const_iterator it=valuesAnswer->values.begin(); it!=valuesAnswer->values.end(); ++it)
				{
					UInt32 localId = it->first;
					UInt32 globalId = getGlobalID(socket, localId);
					r->answer.values[globalId] = it->second;
					r->counter--;
				}

				if (r->counter==0)
				{
					r->answer.serialize(r->client);
					r->client->flush();
					r->answer.deleteValues();
					delete r;
					requests.erase(requestID);
				}

				delete valuesAnswer;
				return;
			}
		}
		else
		{
			// This is data from the clients, so it's a call
			Call *call = Call::deserialize(socket);
			if (socket->isConnected() == false)
				return;

			GetServicesCall *getServicesCall = dynamic_cast<GetServicesCall *>(call);
			if (getServicesCall)
			{
				// We got a GetServicesCall from a client, merge ...
				ServiceAnswer answer;
				for (ExternalServiceMap::const_iterator it=externalServiceMap.begin(); it!=externalServiceMap.end(); ++it)
				{
					Socket *server = it->second.first;
					UInt32 serviceID = it->second.second;
					answer.descriptions[it->first] = serverDescriptionMap[server]->descriptions[serviceID];
				}

				// ... and send
				answer.serialize(socket);
				socket->flush();

				// delete call
				delete getServicesCall;
				return;
			}

			SetValuesCall *setValuesCall = dynamic_cast<SetValuesCall *>(call);
			if (setValuesCall)
			{
				typedef std::map<Socket *, SetValuesCall *> DispatchMap;
				DispatchMap dispatchMap;

				for (std::map<UInt32, ValueVector *>::const_iterator i=setValuesCall->values.begin(); i!=setValuesCall->values.end(); ++i)
				{
					UInt32 globalServiceId = i->first;
					ValueVector *values = i->second;
					if (externalServiceMap.find(globalServiceId) != externalServiceMap.end())
					{
						Socket *server = externalServiceMap[globalServiceId].first;
						UInt32 localServiceId = externalServiceMap[globalServiceId].second;
	
						if (dispatchMap.find(server) == dispatchMap.end())
						{
							// create entry
							dispatchMap[server] = new SetValuesCall;
						}
	
						// add request
						dispatchMap[server]->values[localServiceId] = values;
					}
				}

				// dispatch
				for (DispatchMap::const_iterator i=dispatchMap.begin(); i!=dispatchMap.end(); ++i)
				{
					// send ...
					i->second->serialize(i->first);
					i->first->flush();
					// ... and delete
					delete i->second;
				}

				// delete ValueVectors of caller
				setValuesCall->deleteValues();

				// delete call
				delete setValuesCall;
				return;
			}

			GetValuesCall *getValuesCall = dynamic_cast<GetValuesCall *>(call);
			if (getValuesCall)
			{
				typedef std::map<Socket *, GetValuesCall *> DispatchMap;
				DispatchMap dispatchMap;

				for (unsigned i=0; i<getValuesCall->services.size(); ++i)
				{
					UInt32 globalServiceId = getValuesCall->services[i];
					if (externalServiceMap.find(globalServiceId) != externalServiceMap.end())
					{
						Socket *server = externalServiceMap[globalServiceId].first;
						UInt32 localServiceId = externalServiceMap[globalServiceId].second;
	
						if (dispatchMap.find(server) == dispatchMap.end())
						{
							// create entry
							dispatchMap[server]=new GetValuesCall;
							dispatchMap[server]->requestID = freeRequestId;
						}
	
						// add request
						dispatchMap[server]->services.push_back(localServiceId);
					}
				}

				// if there is nothing to send
				if (dispatchMap.size()==0)
				{
					ValuesAnswer valueAnswer;
					valueAnswer.serialize(socket);
					socket->flush();
					delete getValuesCall;
					return;
				}

				// dispatch
				for (DispatchMap::const_iterator i=dispatchMap.begin(); i!=dispatchMap.end(); ++i)
				{
					// send ...
					i->second->serialize(i->first);
					i->first->flush();
					// ... and delete
					delete i->second;
				}

				// create pending request
				requests[freeRequestId] = new Request(getValuesCall->services.size(), socket);
				freeRequestId++;

				// delete call
				delete getValuesCall;
				return;
			}
		}
	}

	void removeServer(Socket *socket)
	{
		// Erase associations
		for (ExternalServiceMap::iterator it=externalServiceMap.begin(); it!=externalServiceMap.end(); ++it)
		{
			if (it->second.first == socket)
			{
				std::string name = serverDescriptionMap[socket]->descriptions[it->second.second]->name;
				std::cout << "Removing service " << name << std::endl;
				externalServiceMap.erase(it);
				internalServiceMap.erase(it->second);
			}
		}

		// Erase descriptions
		if (serverDescriptionMap[socket])
		{
			serverDescriptionMap[socket]->deleteDescriptions();
			delete serverDescriptionMap[socket];
		}
		serverDescriptionMap.erase(socket);
	}

	void removeSocket(Socket *socket)
	{
		if (serverDescriptionMap.find(socket) != serverDescriptionMap.end())
		{
			removeServer(socket);
		}
		else
		{
			// TODO : do something when removing a client
		}
	}

	void incomingConnection(Socket *socket)
	{
		Call *call = Call::deserialize(socket);
		HelloCall *hello = dynamic_cast<HelloCall *>(call);
		if (hello)
		{
			if (hello->nodeType == HelloCall::SERVER)
			{
				std::cout << "New server connection" << std::endl;
				if (hello->version != Ishtar::PROTOCOL_VERSION)
				{
					std::cerr << "Warning, server " << socket->getRemoteName() << " has not the same Ishtar Protocol Version as proxy" << std::endl;
				}
				
				// We add this server to the server description list
				serverDescriptionMap[socket] = NULL;

				// We request its service description
				GetServicesCall getServicesCall;
				getServicesCall.serialize(socket);
				socket->flush();
			}
			else if (hello->nodeType == HelloCall::CLIENT)
			{
				std::cout << "New client connection" << std::endl;
				if (hello->version != Ishtar::PROTOCOL_VERSION)
				{
					std::cerr << "Warning, client " << socket->getRemoteName() << " has not the same Ishtar Protocol Version as proxy" << std::endl;
				}
			}
			else
				assert(false);
			
			delete call;
		}
		else
		{
			std::cerr << "New connection error : first token was not an hello one" << std::endl;
			socket->disconnect();
		}
	}

	void connectionClosed(Socket *socket)
	{
		std::cout << "Connection closed to " << socket->getRemoteName() << std::endl;
		removeSocket(socket);
	}
};

int main(int argc, char *argv[])
{
	unsigned port = 7000;
	if(argc >= 2)
	{
		port = atoi(argv[1]);
	}

	Proxyd proxyd(port);
	proxyd.run();

	return 0;
}
